// spsh-mit-ls.c
// Hans-Georg Esser, 06.05.2012

#include <sys/stat.h> // lstat()
#include <stdio.h>    // printf()
#include <time.h>     // localtime()
#include <pwd.h>      // getpwuid()
#include <grp.h>      // getgrgid()
#include <stdlib.h>   // exit()
#include <unistd.h>   // exec()
#include <string.h>   // strncpy()
#include <sys/wait.h> // wait()

void ls (char* filename) {
  struct stat s;
  struct passwd *pwd;
  struct group *grp;
  struct tm *time;
  char lnkfilename[280] = "";
  char rights[10];

  if ( lstat (filename, &s) == -1 ) {
    perror("spsh: ls");
    return;
  };
  // printable mode
  rights[0] = '?';
 
  if (S_ISREG(s.st_mode))  rights[0] = '-';
  if (S_ISDIR(s.st_mode))  rights[0] = 'd';
  if (S_ISCHR(s.st_mode))  rights[0] = 'c';
  if (S_ISBLK(s.st_mode))  rights[0] = 'b';
  if (S_ISFIFO(s.st_mode)) rights[0] = 'f';
  if (S_ISLNK(s.st_mode)) {
    rights[0] = 'l';
    strcpy ((char*)&lnkfilename, " ->  ");
    readlink (filename, (char*)(&lnkfilename) + 4 , 256);
  };
  if (S_ISSOCK(s.st_mode)) rights[0] = 's';

  rights[1] = (s.st_mode & 0400) ? 'r' : '-';  // -r-------- ?
  rights[2] = (s.st_mode & 0200) ? 'w' : '-';  // --w------- ?
  rights[3] = (s.st_mode & 0100) ? 'x' : '-';  // ---x------ ?
  rights[4] = (s.st_mode & 0040) ? 'r' : '-';  // ----r----- ?
  rights[5] = (s.st_mode & 0020) ? 'w' : '-';  // -----w---- ?
  rights[6] = (s.st_mode & 0010) ? 'x' : '-';  // ------x--- ?
  rights[7] = (s.st_mode & 0004) ? 'r' : '-';  // -------r-- ?
  rights[8] = (s.st_mode & 0002) ? 'w' : '-';  // --------w- ?
  rights[9] = (s.st_mode & 0001) ? 'x' : '-';  // ---------x ?
  rights[10] = '\0';
  if (s.st_mode & 04000)
    rights[3] = (s.st_mode & 0100) ? 's' : 'S'; // SetUID: 4000, 0100?
  if (s.st_mode & 02000) 
    rights[6] = (s.st_mode & 0010) ? 's' : 'S'; // SetGID: 2000, 0010?
  if (s.st_mode & 01000) 
    rights[9] = (s.st_mode & 0001) ? 't' : 'T'; // sticky: 1000, 0001?

  pwd = getpwuid(s.st_uid);
  grp = getgrgid(s.st_gid);
  time = localtime (&s.st_mtime);

  printf ("%d %s %d %s %s %d %04d/%02d/%02d %02d:%02d:%02d %s%s\n", 
    (int)s.st_ino, rights, (int)s.st_nlink,
    pwd->pw_name, grp->gr_name, (int)s.st_size,
    time->tm_year+1900, time->tm_mon+1, time->tm_mday,
    time->tm_hour, time->tm_min, time->tm_sec, 
    filename, lnkfilename);
  return;
};


int main () {
  char command[255];
  char seps[] = " \t";
  char *part;
  char *args[10];
  short no_args;
  int background, status, pid, wpid;

  while (1) {
    // Beendete Kindprozesse einsammeln
    wpid = waitpid (WAIT_ANY, &status, WNOHANG);
    while ( wpid > 0) {
      printf ("Process %d terminated.\n", wpid); 
      wpid = waitpid (WAIT_ANY, &status, WNOHANG);
    };
  
    printf ("spsh$ ");
    fgets (command, sizeof(command), stdin);
    // Aus Eingabe \n abschneiden
    command[strlen(command)-1] = (char) 0;

    no_args = 0;
    background = 0;
    
    part = strtok (command, seps);
    while ( part != NULL ) {
      args[no_args] = part;
      no_args++;
      part = strtok (NULL, seps);
    };

    // Test, ob letztes Argument ein "&" ist -> background
    if (!strcmp(args[no_args-1],"&")) {
      no_args--;
      background = 1;
    };

    if (no_args == 0) continue;  // kein fork/exec

    // internes Kommando?
    if (!strcmp(args[0],"exit")) {
      printf ("Terminating\n");
      exit(0);
    };
    if (!strcmp(args[0],"ls") && no_args == 2) {
      ls (args[1]);
      continue;
    };
    
    // Kommando mit exec ausfuehren
    args[no_args] = NULL;

    pid = fork();
    if ( pid == 0 ) {
      // Kindprozess
      execvp (args[0], args);
      // exec fehlgeschlagen?
      printf ("%s not found\n", args[0]);
      exit(0);
    } else {
      printf ("Process %d launched: ", pid);
      // Vaterprozess
      if (background == 0) {
        printf ("foreground\n");
        waitpid (pid, &status, 0);
      } else {
        printf ("background\n");
      };
      background = 0;
    };
  }
}

